import hashlib
import hmac
import json
import urllib
from datetime import datetime

from django.contrib.auth.models import Group
from django.core import mail
from django.test import TestCase
from django.utils import timezone
from mock import Mock, patch

from job_runner.apps.job_runner.models import (
    Job,
    JobTemplate,
    KillRequest,
    Project,
    Run,
    RunLog,
    Worker,
    WorkerPool,
)


class ApiTestBase(TestCase):
    """
    Base class for API testing.
    """
    def get(self, path, *args, **kwargs):
        api_key = hmac.new(
            'verysecret', 'GET{0}'.format(path), hashlib.sha1).hexdigest()

        return self.client.get(
            path,
            *args,
            ACCEPT='application/json',
            HTTP_AUTHORIZATION='ApiKey worker1:{0}'.format(api_key)
        )

    def get_json(self, path, *args, **kwargs):
        response = self.get(path, *args, **kwargs)
        return json.loads(response.content)

    def post(self, path, data, method_override='POST'):
        json_data = json.dumps(data)
        api_key = hmac.new(
            'verysecret', '{0}{1}{2}'.format(method_override, path, json_data),
            hashlib.sha1
        ).hexdigest()

        return self.client.post(
            path,
            data=json_data,
            content_type='application/json',
            ACCEPT='application/json',
            HTTP_AUTHORIZATION='ApiKey worker1:{0}'.format(api_key),
            HTTP_X_HTTP_METHOD_OVERRIDE=method_override,
        )

    def patch(self, path, data):
        return self.post(path, data, method_override='PATCH')


class GroupTestCase(ApiTestBase):
    """
    Tests for the group interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_project_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/group/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(['get'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/group/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Group 1', json_data['objects'][0]['name'])

        response = self.get('/api/v1/group/1/')
        self.assertEqual(200, response.status_code)

        response = self.get('/api/v1/group/2/')
        self.assertEqual(401, response.status_code)

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/group/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Group 1', json_data['objects'][0]['name'])

        response = self.client.get(
            '/api/v1/group/1/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        response = self.client.get(
            '/api/v1/group/2/', ACCEPT='application/json')
        self.assertEqual(401, response.status_code)

    def test_filter_on_title(self):
        """
        Test filtering on project title.
        """
        response = self.get('/api/v1/project/?{0}'.format(
            urllib.urlencode({'title': 'Test project 1'})))
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, json_data['objects'][0]['id'])


class ProjectTestCase(ApiTestBase):
    """
    Tests for the project interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_project_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/project/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(['get'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_projects_count(self):
        """
        Test that we have two projects in the database.
        """
        self.assertEqual(2, Project.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/project/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test project 1', json_data['objects'][0]['title'])

        response = self.get('/api/v1/project/1/')
        self.assertEqual(200, response.status_code)

        response = self.get('/api/v1/project/2/')
        self.assertEqual(401, response.status_code)

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/project/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test project 1', json_data['objects'][0]['title'])

        response = self.client.get(
            '/api/v1/project/1/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        response = self.client.get(
            '/api/v1/project/2/', ACCEPT='application/json')
        self.assertEqual(401, response.status_code)


class WorkerPoolTestCase(ApiTestBase):
    """
    Tests for the worker-pool interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_worker_pool_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/worker_pool/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(['get'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_worker_pool_count(self):
        """
        Test worker-pool count in the DB.
        """
        self.assertEqual(2, WorkerPool.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/worker_pool/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Pool 1', json_data['objects'][0]['title'])

        response = self.get('/api/v1/worker_pool/1/')
        self.assertEqual(200, response.status_code)

        response = self.get('/api/v1/worker_pool/2/')
        self.assertEqual(401, response.status_code)

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/worker_pool/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Pool 1', json_data['objects'][0]['title'])

        response = self.client.get(
            '/api/v1/worker_pool/1/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        response = self.client.get(
            '/api/v1/worker_pool/2/', ACCEPT='application/json')
        self.assertEqual(401, response.status_code)

    def test_filter_on_title(self):
        """
        Test filter on worker-pool title.
        """
        response = self.get('/api/v1/worker_pool/?{0}'.format(
            urllib.urlencode({'title': 'Pool 1'})))
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, json_data['objects'][0]['id'])


class WorkerTestCase(ApiTestBase):
    """
    Tests for the worker interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_worker_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/worker/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(
            ['get', 'patch'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_worker_count(self):
        """
        Test worker count in the DB.
        """
        self.assertEqual(2, Worker.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/worker/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test worker 1', json_data['objects'][0]['title'])

        response = self.get('/api/v1/worker/1/')
        self.assertEqual(200, response.status_code)

        response = self.get('/api/v1/worker/2/')
        self.assertEqual(401, response.status_code)

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/worker/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test worker 1', json_data['objects'][0]['title'])

        response = self.client.get(
            '/api/v1/worker/1/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        response = self.client.get(
            '/api/v1/worker/2/', ACCEPT='application/json')
        self.assertEqual(401, response.status_code)

    def test_patch_ping_response_dts(self):
        """
        Test PATCH the ping_response_dts field.
        """
        dts = timezone.now()
        response = self.patch(
            '/api/v1/worker/1/',
            {'ping_response_dts': dts.isoformat(' ')}
        )
        self.assertEqual(202, response.status_code)

        worker = Worker.objects.get(pk=1)
        self.assertEqual(dts, worker.ping_response_dts)

    def test_patch_ping_response_no_permissions(self):
        """
        Test PATCH the ping_response_dts without having permission.
        """
        dts = timezone.now()
        response = self.patch(
            '/api/v1/worker/2/',
            {'ping_response_dts': dts.isoformat(' ')}
        )
        self.assertEqual(401, response.status_code)


class JobTemplateTestCase(ApiTestBase):
    """
    Tests for the job-template interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_job_template_methods(self):
        """
        Test allowed methods.
        """
        response = self.get('/api/v1/job_template/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(['get'], json_data['allowed_detail_http_methods'])
        self.assertEqual(['get'], json_data['allowed_list_http_methods'])

    def test_job_template_count(self):
        """
        Test job-template count in the db.
        """
        self.assertEqual(2, JobTemplate.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/job_template/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test template 1', json_data['objects'][0]['title'])

        response = self.get('/api/v1/job_template/1/')
        self.assertEqual(200, response.status_code)

        response = self.get('/api/v1/job_template/2/')
        self.assertEqual(401, response.status_code)

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/job_template/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test template 1', json_data['objects'][0]['title'])

        response = self.client.get(
            '/api/v1/job_template/1/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        response = self.client.get(
            '/api/v1/job_template/2/', ACCEPT='application/json')
        self.assertEqual(401, response.status_code)

    def test_filter_on_title(self):
        """
        Test filtering a job-template on title.
        """
        response = self.get('/api/v1/job_template/?{0}'.format(
            urllib.urlencode({'title': 'Test template 1'})))
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, json_data['objects'][0]['id'])


class JobTestCase(ApiTestBase):
    """
    Tests for the job interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_job_methods(self):
        """
        Test allowed methods on jobs.
        """
        response = self.get('/api/v1/job/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertItemsEqual(
            ['get', 'put', 'patch'],
            json_data['allowed_detail_http_methods']
        )
        self.assertItemsEqual(
            ['get', 'post'],
            json_data['allowed_list_http_methods']
        )

    def test_job_count(self):
        """
        Test the job count in the db.
        """
        self.assertEqual(2, Job.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/job/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test job 1', json_data['objects'][0]['title'])

        response = self.get('/api/v1/job/1/')
        self.assertEqual(200, response.status_code)

        response = self.get('/api/v1/job/2/')
        self.assertEqual(401, response.status_code)

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/job/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual('Test job 1', json_data['objects'][0]['title'])

        response = self.client.get(
            '/api/v1/job/1/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        response = self.client.get(
            '/api/v1/job/2/', ACCEPT='application/json')
        self.assertEqual(401, response.status_code)

    def test_filter_on_title(self):
        """
        Test filtering on job-title.
        """
        response = self.get('/api/v1/job/?{0}'.format(urllib.urlencode({
            'title': 'Test job 1',
        })))
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, json_data['objects'][0]['id'])


class RunTestCase(ApiTestBase):
    """
    Tests for the run interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_run_methods(self):
        """
        Test allowed methods on runs.
        """
        response = self.get('/api/v1/run/schema/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual([
            'get',
            'patch',
        ], json_data['allowed_detail_http_methods'])
        self.assertEqual(
            ['get', 'post'], json_data['allowed_list_http_methods'])

    def test_run_count(self):
        """
        Test the run count in the db.
        """
        self.assertEqual(2, Run.objects.count())

    def test_api_authorization(self):
        """
        Test API authorization (API-key has access to one object).
        """
        response = self.get('/api/v1/run/')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual(1, json_data['objects'][0]['id'])

        response = self.get('/api/v1/run/1/')
        self.assertEqual(200, response.status_code)

        response = self.get('/api/v1/run/2/')
        self.assertEqual(401, response.status_code)

    def test_user_authorization(self):
        """
        Test user authorization (user has only access to one object).
        """
        self.client.login(username='admin', password='admin')
        response = self.client.get(
            '/api/v1/run/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        json_data = json.loads(response.content)
        self.assertEqual(1, len(json_data['objects']))
        self.assertEqual(1, json_data['objects'][0]['id'])

        response = self.client.get(
            '/api/v1/run/1/', ACCEPT='application/json')
        self.assertEqual(200, response.status_code)

        response = self.client.get(
            '/api/v1/run/2/', ACCEPT='application/json')
        self.assertEqual(401, response.status_code)

    def test_scheduled(self):
        """
        Test listing scheduled runs.
        """
        expected = [
            ('scheduled', 1),
            ('in_queue', 0),
            ('started', 0),
            ('completed', 0),
            ('completed_successful', 0),
            ('completed_with_error', 0),
        ]

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_in_queue(self):
        """
        Test listing runs in queue.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 1),
            ('started', 0),
            ('completed', 0),
            ('completed_successful', 0),
            ('completed_with_error', 0),
        ]

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.save()

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_started(self):
        """
        Test listing runs that are started.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 0),
            ('started', 1),
            ('completed', 0),
            ('completed_successful', 0),
            ('completed_with_error', 0),
        ]

        run = Run.objects.get(pk=1)
        run.enqueue_dts = timezone.now()
        run.start_dts = timezone.now()
        run.save()

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    @patch('job_runner.apps.job_runner.signals.post_run_update', Mock())
    def test_completed(self):
        """
        Test listing runs that are completed.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 0),
            ('started', 0),
            ('completed', 1),
            ('completed_successful', 1),
            ('completed_with_error', 0),
        ]

        Run.objects.filter(pk=1).update(
            enqueue_dts=timezone.now(),
            start_dts=timezone.now(),
            return_dts=timezone.now(),
            return_success=True,
        )

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_completed_with_error(self):
        """
        Test listing runs that are completed with error.
        """
        expected = [
            ('scheduled', 0),
            ('in_queue', 0),
            ('started', 0),
            ('completed', 1),
            ('completed_successful', 0),
            ('completed_with_error', 1),
        ]

        Run.objects.filter(pk=1).update(
            enqueue_dts=timezone.now(),
            start_dts=timezone.now(),
            return_dts=timezone.now(),
            return_success=False
        )

        for argument, expected in expected:
            json_data = self.get_json(
                '/api/v1/run/?state={0}'.format(argument))
            self.assertEqual(expected, len(json_data['objects']))

    def test_patch_run(self):
        """
        Test PATCH ``/api/v1/run/1/``.
        """
        enqueue_dts = timezone.now()

        response = self.patch(
            '/api/v1/run/1/',
            {
                'enqueue_dts': enqueue_dts.isoformat(' ')
            }
        )

        self.assertEqual(202, response.status_code)
        run = Run.objects.get(pk=1)
        self.assertEqual(enqueue_dts, run.enqueue_dts)

    def test_patch_run_no_permission(self):
        """
        Test PATCH ``/api/v1/run/2/`` (we don't have access to this object).
        """
        enqueue_dts = timezone.now()

        response = self.patch(
            '/api/v1/run/2/',
            {
                'enqueue_dts': enqueue_dts.isoformat(' ')
            }
        )

        self.assertEqual(401, response.status_code)

    def test_returned_with_error(self):
        """
        Test PATCH ``/api/v1/run/1/`` returned with error.

        This is expected to send e-mail notifications out.

        """
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': False,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertTrue(hasattr(mail, 'outbox'))
        self.assertEqual(1, len(mail.outbox))
        self.assertEqual(4, len(mail.outbox[0].to))
        self.assertEqual('Run error for: Test job 1', mail.outbox[0].subject)
        self.assertEqual(1, Job.objects.get(pk=1).last_completed_schedule_id)

    def test_returned_with_error_disable_enqueue(self):
        """
        Test PATCH ``/api/v1/run/1/`` error disables enqueue.

        This is expected to disable the enqueue.

        """
        job = Job.objects.get(pk=1)
        job.disable_enqueue_after_fails = 3
        job.save()

        for i in range(3):
            response = self.patch(
                '/api/v1/run/1/',
                {
                    'return_dts': timezone.now().isoformat(' '),
                    'return_success': False,
                }
            )

            self.assertEqual(202, response.status_code)
            job = Job.objects.get(pk=1)
            self.assertTrue(job.enqueue_is_enabled)
            self.assertEqual(i + 1, job.fail_times)

        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': False,
            }
        )

        job = Job.objects.get(pk=1)
        self.assertFalse(job.enqueue_is_enabled)
        self.assertEqual(1, job.last_completed_schedule_id)

    def test_create_new_run(self):
        """
        Test create a new run by performing a ``POST``.
        """
        self.client.login(username='admin', password='admin')

        response = self.client.post(
            '/api/v1/run/',
            data=json.dumps({
                'job': '/api/v1/job/1/',
                'schedule_dts': '2013-01-01 00:00:00',
            }),
            content_type='application/json',
            ACCEPT='application/json',
        )

        self.assertEqual(201, response.status_code)
        run = Run.objects.filter(job_id=1)[1]
        self.assertEqual(
            timezone.make_aware(
                datetime(2013, 1, 1),
                timezone.get_default_timezone()
            ),
            run.schedule_dts
        )

    def test_create_new_run_no_permission(self):
        """
        Test create a new run by performing a ``POST`` with invalid group.
        """
        project = Project.objects.get(pk=1)
        project.auth_groups = [Group.objects.get(pk=2)]
        project.save()

        self.client.login(username='admin', password='admin')

        response = self.client.post(
            '/api/v1/run/',
            data=json.dumps({
                'job': '/api/v1/job/1/',
                'schedule_dts': '2013-01-01 00:00:00',
            }),
            content_type='application/json',
            ACCEPT='application/json',
        )

        self.assertEqual(401, response.status_code)


class ChainedRunTestCase(ApiTestBase):
    """
    Tests for the run interface with chaining.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
        'test_child_job',
    ]

    def test_patch_with_reschedule(self):
        """
        Test PATCH ``/api/v1/run/1/`` for chained job.

        """
        Run.objects.update(enqueue_dts=timezone.now())
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': True,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(1, Job.objects.get(pk=3).run_set.count())

    def test_patch_failed_run(self):
        """
        Test PATCH ``/api/v1/run/1/`` and mark run as failed.

        In this case, we're not expecting the job to schedule it's child.

        """
        Run.objects.update(enqueue_dts=timezone.now())
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': False,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())

    def test_patch_failed_run_with_schedule_child(self):
        """
        Test PATCH ``/api/v1/run/1/`` as failed, but schedule child.

        Since we've set ``schedule_children_on_error=True``, we expect that
        the chain will continue, even when the job failed.

        """
        Run.objects.update(enqueue_dts=timezone.now())
        Job.objects.filter(pk=1).update(schedule_children_on_error=True)
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': False,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(1, Job.objects.get(pk=3).run_set.count())

    def test_patch_with_disabled_child(self):
        """
        Test PATCH ``/api/v1/run/1/`` for chained job with disabled child.

        When enqueue_is_enabled=False, we do not expect that the child will
        be scheduled.

        """
        Job.objects.filter(pk=3).update(enqueue_is_enabled=False)
        Run.objects.update(enqueue_dts=timezone.now())
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': True,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())

    def test_patch_with_reschedule_no_children_reschedule(self):
        """
        Test PATCH ``/api/v1/run/1/`` for chained job but not reschedule
        children.

        """
        Run.objects.update(
            enqueue_dts=timezone.now(),
            schedule_children=False,
        )
        response = self.patch(
            '/api/v1/run/1/',
            {
                'return_dts': timezone.now().isoformat(' '),
                'return_success': True,
            }
        )

        self.assertEqual(202, response.status_code)
        self.assertEqual(0, Job.objects.get(pk=3).run_set.count())


class KillRequestTestCase(ApiTestBase):
    """
    Tests for the kill request interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_post_new_kill_request(self):
        """
        Test POST ``/api/v1/kill_request/``.
        """
        response = self.post(
            '/api/v1/kill_request/',
            {
                'run': '/api/v1/run/1/',
            }
        )

        self.assertEqual(201, response.status_code)
        self.assertEqual(1, KillRequest.objects.filter(run_id=1).count())

    def test_post_new_kill_request_no_permission(self):
        """
        Test POST ``/api/v1/kill_request/`` with run we don't have access to.
        """
        response = self.post(
            '/api/v1/kill_request/',
            {
                'run': '/api/v1/run/2/',
            }
        )
        self.assertEqual(401, response.status_code)


class RunLogTestCase(ApiTestBase):
    """
    Tests for the run log interface.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test_post_new_run_log(self):
        """
        Test POST ``/api/v1/run_log/``.
        """
        response = self.post(
            '/api/v1/run_log/',
            {
                'run': '/api/v1/run/1/',
                'content': 'foo bar',
            }
        )
        self.assertEqual(201, response.status_code)
        self.assertEqual(1, RunLog.objects.filter(run_id=1).count())

    def test_post_new_run_log_no_access(self):
        """
        Test POST ``/api/v1/run_log/`` with run we don't have access to.
        """
        response = self.post(
            '/api/v1/run_log/',
            {
                'run': '/api/v1/run/2/',
                'content': 'foo bar',
            }
        )
        self.assertEqual(401, response.status_code)

    def test_update_log(self):
        """
        Test update a log item.
        """
        response = self.post(
            '/api/v1/run_log/',
            {
                'run': '/api/v1/run/1/',
                'content': 'foo bar',
            }
        )
        self.assertEqual(201, response.status_code)

        response = self.patch(
            '/api/v1/run_log/1/',
            {
                'content': 'bar foo',
            }
        )
        self.assertEqual(202, response.status_code)



import json
from datetime import timedelta

from django.core import mail
from django.test import TestCase
from django.utils import timezone
from mock import Mock

from job_runner.apps.job_runner.management.commands.health_check import Command
from job_runner.apps.job_runner.models import Run, Worker


class CommandTestCase(TestCase):
    """
    Tests for :class:`.Command`.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test__mark_worker_runs_as_failed(self):
        """
        Test :meth:`.Command._mark_worker_runs_as_failed`.
        """
        worker = Worker.objects.get(pk=1)
        Run.objects.filter(pk=1).update(worker=worker)

        command = Command()
        command.publisher = Mock()

        self.assertEqual(None, Run.objects.get(pk=1).return_dts)

        command._mark_worker_runs_as_failed(worker, 'Test mark as failed')

        run = Run.objects.get(pk=1)
        self.assertNotEqual(None, run.return_dts)
        self.assertFalse(run.return_success)

        command.publisher.send_multipart.assert_called_once_with([
            'worker.event',
            json.dumps({
                'event': 'returned',
                'run_id': 1,
                'kind': 'run',
            })
        ])

    def test__find_unresponsive_workers_and_mark_runs_as_failed(self):
        """
        Test ``_find_unresponsive_workers_and_mark_runs_as_failed``.
        """
        Run.objects.filter(pk=1).update(worker=Worker.objects.get(pk=1))
        Run.objects.filter(pk=2).update(worker=Worker.objects.get(pk=2))

        # In testing.py:
        # JOB_RUNNER_WORKER_PING_INTERVAL = 60 * 5
        # JOB_RUNNER_WORKER_MARK_JOB_FAILED_AFTER_INTERVALS = 3
        acceptable = timezone.now() - timedelta(seconds=(60 * 5 * 3))
        unacceptable = acceptable - timedelta(seconds=15)

        Worker.objects.filter(pk=1).update(ping_response_dts=acceptable)
        Worker.objects.filter(pk=2).update(ping_response_dts=unacceptable)

        command = Command()
        command.publisher = Mock()

        command._find_unresponsive_workers_and_mark_runs_as_failed()

        runs = Run.objects.all()

        # Run pk=1 was marked as failed
        self.assertNotEqual(None, runs[0].return_dts)
        self.assertFalse(runs[0].return_success)

        # Run pk=2 was not touched
        self.assertEqual(None, runs[1].return_success)
        self.assertEqual(None, runs[1].return_dts)

    def test__find_unresponsive_worker_pools(self):
        """
        Test :meth:`.Command._find_unresponsive_worker_pools`.

        In this case, we expect Pool 1 to be responsive and Pool 2 to be
        unresponsive.

        """
        Worker.objects.filter(pk=1).update(ping_response_dts=timezone.now())
        # from settings in ``base.py``:
        # 915 = (60 * 5 * 3) + 15
        Worker.objects.filter(pk=2).update(
            ping_response_dts=timezone.now() - timedelta(seconds=915))

        command = Command()
        command._find_unresponsive_worker_pools()

        self.assertEqual(1, len(mail.outbox))
        self.assertEqual(
            'Worker-pool unresponsive: Pool 2', mail.outbox[0].subject)



import json
import logging
import time
from datetime import datetime, timedelta

import zmq
from django.conf import settings
from django.core.management.base import NoArgsCommand
from django.db import transaction

from job_runner.apps.job_runner.models import Worker, WorkerPool
from job_runner.apps.job_runner import notifications


logger = logging.getLogger(__name__)


class Command(NoArgsCommand):
    help = 'Monitor worker and worker-pool health'

    publisher = None
    """
    Holds the ZMQ ``publisher`` instance, used to publish to the WS server.
    """

    @transaction.commit_manually
    def handle_noargs(self, **options):
        logger.info('Starting health check monitor')
        context = zmq.Context(1)

        # setup the event publisher which is publishing to the WS server
        self.publisher = context.socket(zmq.PUB)
        self.publisher.connect('tcp://{0}:{1}'.format(
            settings.JOB_RUNNER_WS_SERVER_HOSTNAME,
            settings.JOB_RUNNER_WS_SERVER_PORT,
        ))

        # give the subscribers some time to (re-)connect.
        time.sleep(2)

        hc_delta = timedelta(
            seconds=settings.JOB_RUNNER_WORKER_HEALTH_CHECK_INTERVAL)
        next_hc = datetime.utcnow() + hc_delta

        while True:
            if next_hc <= datetime.utcnow():
                self._find_unresponsive_workers_and_mark_runs_as_failed()
                self._find_unresponsive_worker_pools()
                next_hc = datetime.utcnow() + hc_delta
            transaction.commit()
            time.sleep(5)

        self.publisher.close()
        context.term()

    def _find_unresponsive_workers_and_mark_runs_as_failed(self):
        """
        Detect unresponsive workers and mark their active runs as failed.
        """
        logger.info('Looking for unresponsive workers.')
        workers = Worker.objects.filter(enqueue_is_enabled=True)

        for worker in workers:
            if not worker.is_responsive():
                logger.warning(
                    'Worker ID {0} seems unresponsive.'.format(worker.pk))
                self._mark_worker_runs_as_failed(
                    worker,
                    'Worker does not respond to ping requests.'
                )

    def _find_unresponsive_worker_pools(self):
        """
        Detect worker-pools where all workers are unresponsive.

        In this case, an e-mail will be send with a warning.

        """
        logger.info('Looking for unresponsive worker-pools.')
        worker_pools = WorkerPool.objects.filter(enqueue_is_enabled=True)

        for worker_pool in worker_pools:
            is_responsive = False
            workers = worker_pool.workers.filter(enqueue_is_enabled=True)
            for worker in workers:
                if worker.is_responsive():
                    is_responsive = True

            if not is_responsive and workers.count():
                notifications.worker_pool_unresponsive(worker_pool)

    @transaction.commit_manually
    def _mark_worker_runs_as_failed(self, worker, message):
        """
        Mark the runs that are not finished yet as failed for ``worker``.

        :param worker:
            The worker for which to mark the runs as failed.

        :param message:
            A ``str`` containing a message why the runs were marked as failed.

        """
        run_ids = []

        try:
            active_runs = worker.run_set.filter(return_dts__isnull=True)
            active_runs = active_runs.select_for_update()

            for run in active_runs:
                logger.warning('Marking run {0} as failed.'.format(run.pk))
                run.mark_failed(message)
                run_ids.append(run.pk)

        except Exception:
            logger.exception('Something went wrong, rolling back transaction')
            transaction.rollback()
        else:
            transaction.commit()

        # broadcast events after transaction has been comitted :-)
        for run_id in run_ids:
            self.publisher.send_multipart([
                'worker.event',
                json.dumps({
                    'event': 'returned',
                    'run_id': run_id,
                    'kind': 'run',
                })
            ])



from datetime import datetime, timedelta

from django.test import TestCase
from django.utils.timezone import get_current_timezone

from job_runner.apps.job_runner.utils import correct_dst_difference


class ModuleTestCase(TestCase):
    """
    Tests for :mod:`~job_runner.apps.job_runner.utils`.
    """
    def test_correct_dts_difference(self):
        """
        Test :func:`.correct_dst_difference`.

        This will test that during a DST change (from +1 to +2), the
        :func:`.correct_dst_difference` function will make sure that the
        local will stay the same.

        """
        tzinfo = get_current_timezone()
        prev_dts = tzinfo.localize(datetime(2013, 3, 30, 4))
        next_dts = prev_dts + timedelta(days=1)
        expected_next_dts = tzinfo.localize(datetime(2013, 3, 31, 4))

        # test that indeed there is a DST change
        self.assertEqual(timedelta(hours=1), prev_dts.utcoffset())
        self.assertEqual(timedelta(hours=2), expected_next_dts.utcoffset())

        # test that normally, the job would be scheduled at 05:00
        self.assertEqual(
            tzinfo.localize(datetime(2013, 3, 31, 5)),
            next_dts
        )

        # test that correct_dst_difference is correcting this to 04:00
        self.assertEqual(
            expected_next_dts,
            correct_dst_difference(prev_dts, next_dts)
        )



from django.utils.timezone import get_current_timezone


def correct_dst_difference(previous_schedule_dts, next_schedule_dts):
    """
    Calculate the daylight saving-time difference between from and to DTS.

    Example: You are in Europe/Amsterdam timezone with DST (GMT+2) and you have
    a daily job running at 10:00 each morning. This translates to 08:00 UTC
    time. When switching to winter time (GMT+1), this would mean that the
    job suddenly starts at 09:00 (08:00 UTC + 1).

    By calculating the UTC offset between the previous schedule DTS and the
    next schedule DTS, we can find out if there is a DST change, and compensate
    this difference, to make sure the job will be re-scheduled at the same
    time in the local timezone.

    """
    previous_dts = previous_schedule_dts.astimezone(get_current_timezone())
    next_dts = next_schedule_dts.astimezone(get_current_timezone())

    dts_difference = previous_dts.utcoffset() - next_dts.utcoffset()

    return next_dts + dts_difference



import copy

from django.conf import settings
from django.core.mail import send_mail
from django.template import Context
from django.template.loader import get_template


def reschedule_failed(job):
    """
    Send out a notification that the given ``job`` failed to reschedule.
    """
    t = get_template('job_runner/email/reschedule_failed.txt')
    c = Context({
        'job': job,
        'hostname': settings.HOSTNAME,
    })
    email_body = t.render(c)

    addresses = copy.copy(settings.JOB_RUNNER_ADMIN_EMAILS)
    addresses.extend(job.get_notification_addresses())

    if addresses:
        send_mail(
            'Reschedule error for: {0}'.format(job.title),
            email_body,
            settings.DEFAULT_FROM_EMAIL,
            addresses
        )


def run_failed(run):
    """
    Send out a notification that the given ``run`` failed.
    """
    t = get_template('job_runner/email/job_failed.txt')
    c = Context({
        'time_zone': settings.TIME_ZONE,
        'run': run,
        'hostname': settings.HOSTNAME,
    })
    email_body = t.render(c)

    addresses = copy.copy(settings.JOB_RUNNER_ADMIN_EMAILS)
    addresses.extend(run.job.get_notification_addresses())

    if addresses:
        send_mail(
            'Run error for: {0}'.format(run.job.title),
            email_body,
            settings.DEFAULT_FROM_EMAIL,
            addresses
        )


def worker_pool_unresponsive(worker_pool):
    """
    Send out a notification that the given ``worker_pool`` is unresponsive.
    """
    t = get_template('job_runner/email/worker_pool_unresponsive.txt')
    c = Context({
        'worker_pool': worker_pool,
        'hostname': settings.HOSTNAME,
    })
    email_body = t.render(c)

    addresses = copy.copy(settings.JOB_RUNNER_ADMIN_EMAILS)
    addresses.extend(worker_pool.get_notification_addresses())

    if addresses:
        send_mail(
            'Worker-pool unresponsive: {0}'.format(worker_pool.title),
            email_body,
            settings.DEFAULT_FROM_EMAIL,
            addresses
        )



from datetime import datetime, time, timedelta

from django.core import mail
from django.core.exceptions import ValidationError
from django.test import TestCase
from django.utils import timezone

from job_runner.apps.job_runner.models import (
    Job,
    RescheduleExclude,
    Run,
)
from job_runner.apps.job_runner.utils import correct_dst_difference


class JobTestCase(TestCase):
    """
    Tests for the job model.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs'
    ]

    def test_recursion(self):
        """
        Test that saving a resursive job-chain will raise a validation error.
        """
        job = Job.objects.get(pk=2)
        job.parent = Job.objects.get(pk=1)
        job.full_clean()
        job.save()

        job = Job.objects.get(pk=1)
        job.full_clean()
        job.parent = Job.objects.get(pk=2)
        self.assertRaises(ValidationError, job.full_clean)


class RunTestCase(TestCase):
    """
    Tests for the job run model.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs'
    ]

    def test_reschedule(self):
        """
        Test reschedule.
        """
        self.assertEqual(1, Run.objects.filter(job_id=1).count())
        Job.objects.get(pk=1).reschedule()
        # because of the pre-scheduling, reschedule should result in max
        # 2 scheduled runs.
        self.assertEqual(2, Run.objects.filter(job_id=1).count())
        Job.objects.get(pk=1).reschedule()
        self.assertEqual(2, Run.objects.filter(job_id=1).count())

        Run.objects.filter(pk=1).update(
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
            return_dts=timezone.now()
        )

        # we just completed one run, so reschedule should result in a new
        # schedule.
        Job.objects.get(pk=1).reschedule()
        self.assertEqual(3, Run.objects.filter(job_id=1).count())

        # make sure the last run is scheduled based on the schedule_dts of the
        # pre-last run
        runs = Run.objects.filter(job_id=1).all()
        self.assertEqual(
            correct_dst_difference(
                runs[1].schedule_dts,
                runs[1].schedule_dts + timedelta(days=1)
            ),
            runs[2].schedule_dts
        )

    def test_reschedule_not_in_past(self):
        """
        Test reschedule dts is never in the past.
        """
        dts_now = timezone.now()

        Run.objects.filter(pk=1).update(
            schedule_dts=dts_now - timedelta(days=31),
            enqueue_dts=dts_now - timedelta(days=31),
            return_dts=dts_now - timedelta(days=31)
        )

        Job.objects.get(pk=1).reschedule()
        self.assertEqual(2, Run.objects.filter(job_id=1).count())

        run = Run.objects.get(pk=3)

        # we have to make sure to use correct_dst_difference since between
        # 31 days ago and tomorrow, we might have changed from or to DTS.
        self.assertEqual(
            run.schedule_dts,
            correct_dst_difference(
                dts_now - timedelta(days=31),
                dts_now + timedelta(days=1)
            )
        )

    def test_reschedule_monthly_after_schedule_dts(self):
        """
        Test monthy reschedule after schedule dts.
        """
        in_and_expected = [
            (datetime(2032, 1, 1), datetime(2032, 2, 1)),
            (datetime(2032, 2, 1), datetime(2032, 3, 1)),
            (datetime(2032, 3, 1), datetime(2032, 4, 1)),
            (datetime(2032, 4, 1), datetime(2032, 5, 1)),
            (datetime(2032, 6, 1), datetime(2032, 7, 1)),
            (datetime(2032, 12, 1), datetime(2033, 1, 1)),
        ]

        for in_dts, expected_dts in in_and_expected:

            # if you do not use utc as a timezone here, you get funny results
            # because of day light saving time switching
            in_dts = timezone.make_aware(
                in_dts, timezone.utc)
            expected_dts = timezone.make_aware(
                expected_dts, timezone.utc)

            job = Job.objects.get(pk=1)
            job.reschedule_interval_type = 'MONTH'
            job.save()

            job.run_set.all().delete()

            Run.objects.create(
                job=job,
                schedule_dts=in_dts,
                enqueue_dts=in_dts,
                start_dts=in_dts,
                return_dts=in_dts,
            )

            job.reschedule()
            run = Run.objects.filter(job=1, enqueue_dts__isnull=True)[0]
            self.assertEqual(
                correct_dst_difference(in_dts, expected_dts),
                run.schedule_dts
            )

    def test_reschedule_with_exclude(self):
        """
        Test reschedule with exclude time.
        """
        job = Job.objects.get(pk=1)
        job.reschedule_interval_type = 'HOUR'
        job.save()

        Run.objects.filter(pk=1).update(
            enqueue_dts=timezone.now(),
            schedule_dts=timezone.make_aware(
                datetime(2032, 1, 1, 11, 59), timezone.get_default_timezone()),
            return_dts=timezone.make_aware(
                datetime(2032, 1, 1, 11, 59), timezone.get_default_timezone()),
        )

        RescheduleExclude.objects.create(
            job=job,
            start_time=time(12, 00),
            end_time=time(13, 00),
        )

        job.reschedule()

        self.assertEqual(2, Run.objects.filter(job_id=1).count())

        runs = Run.objects.filter(job_id=1)
        self.assertEqual(
            timezone.make_aware(
                datetime(2032, 1, 1, 13, 59),
                timezone.get_default_timezone()
            ),
            runs[1].schedule_dts
        )

    def test_reschedule_with_invalid_exclude(self):
        """
        Test reschedule with exclude time which is invalid.
        """
        job = Job.objects.get(pk=1)
        job.reschedule_interval_type = 'HOUR'
        job.save()

        Run.objects.filter(pk=1).update(
            enqueue_dts=timezone.now(),
            return_dts=timezone.make_aware(
                datetime(2012, 1, 1, 11, 59), timezone.get_default_timezone())
        )

        RescheduleExclude.objects.create(
            job=job,
            start_time=time(0, 0),
            end_time=time(23, 59),
        )

        job.reschedule()

        self.assertEqual(1, Run.objects.filter(job_id=1).count())
        self.assertTrue(hasattr(mail, 'outbox'))
        self.assertEqual(1, len(mail.outbox))
        self.assertEqual(4, len(mail.outbox[0].to))
        self.assertEqual(
            'Reschedule error for: Test job 1', mail.outbox[0].subject)

    def test_reschedule_with_run_scheduled(self):
        """
        Test reschedule when there are already two runs scheduled.
        """
        job = Job.objects.get(pk=1)
        self.assertEqual(1, Run.objects.filter(job=job).count())

        Run.objects.filter(pk=1).update(
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
            return_dts=timezone.now()
        )

        for x in range(2):
            Run.objects.create(
                job=job,
                schedule_dts=timezone.now()
            )

        self.assertEqual(3, Run.objects.filter(job=job).count())
        job.reschedule()
        self.assertEqual(3, Run.objects.filter(job=job).count())

    def test_reschedule_with_started_run(self):
        """
        Test reschedule when there is an other run started and one scheduled.
        """
        job = Job.objects.get(pk=1)
        self.assertEqual(1, Run.objects.filter(job=job).count())

        Run.objects.filter(pk=1).update(
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
            start_dts=timezone.now(),
            return_dts=timezone.now()
        )

        Run.objects.create(
            job=job,
            schedule_dts=timezone.now()
        )

        Run.objects.create(
            job=job,
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
            start_dts=timezone.now(),
        )

        self.assertEqual(3, Run.objects.filter(job=job).count())
        job.reschedule()
        self.assertEqual(3, Run.objects.filter(job=job).count())

    def test_get_notification_addresses(self):
        """
        Test ``get_notification_addresses`` methods on models.
        """
        self.assertItemsEqual(
            [
                'project1@example.com',
                'pool1@example.com',
                'template1@example.com',
                'job1@example.com',
            ],
            Job.objects.get(pk=1).get_notification_addresses()
        )

    def test_schedule(self):
        """
        Test direct schedule.
        """
        Run.objects.all().delete()

        job = Job.objects.get(pk=1)

        self.assertEqual(0, job.run_set.count())
        job.schedule()
        self.assertEqual(1, job.run_set.count())

    def test_schedule_id_set_on_create(self):
        """
        Test that the ``schedule_id`` is set on create.
        """
        run = Run.objects.create(
            job=Job.objects.get(pk=1),
            schedule_dts=timezone.now()
        )
        self.assertTrue(run.pk == run.schedule_id)

    def test_mark_failed(self):
        """
        Test :meth:`.Run.mark_failed`.
        """
        run = Run.objects.get(pk=1)
        run.mark_failed('Test mark failed')

        run = Run.objects.get(pk=1)

        self.assertEqual(
            'This run was marked as failed. Reason: Test mark failed',
            run.run_log.content
        )
        self.assertIsInstance(run.enqueue_dts, datetime)
        self.assertIsInstance(run.start_dts, datetime)
        self.assertIsInstance(run.return_dts, datetime)
        self.assertFalse(run.return_success)

    def test_mark_failed_on_completed_run(self):
        """
        Test :meth:`.Run.mark_failed` on completed run.

        In this case, :meth:`.Run.mark_failed` shouldn't do anything.

        """
        Run.objects.filter(pk=1).update(
            return_dts=timezone.now(),
            return_success=True,
        )
        run = Run.objects.get(pk=1)
        run.mark_failed('Test mark failed')
        run = Run.objects.get(pk=1)

        self.assertTrue(run.return_success)



from job_runner.apps.job_runner import notifications


def post_run_create(sender, instance, raw, **kwargs):
    """
    Post action after creating a run instance.

    This will make sure the ``schedule_id`` field is set.

    """
    if not instance.schedule_id:
        instance.schedule_id = instance.id
        instance.save()


def post_run_update(sender, instance, created, raw, **kwargs):
    """
    Post action after saving a run instance.

    This will:

        * send an error notification by e-mail when the job failed
        * disable the job when it failed more times than allowed
        * schedule it's children (if applicable)

    """
    if created or raw:
        return

    job = instance.job

    if instance.return_dts:
        if instance.return_success is False:
            # the run failed
            notifications.run_failed(instance)
            job.fail_times += 1
            job.last_completed_schedule_id = instance.schedule_id

            # disable job when it failed more than x times
            if (job.disable_enqueue_after_fails and
                    job.fail_times > job.disable_enqueue_after_fails):
                job.enqueue_is_enabled = False

            job.save()

        else:
            # reset the fail count
            job.fail_times = 0
            job.last_completed_schedule_id = instance.schedule_id
            job.save()

        # on purpose we are not using .count() since with that, the
        # .select_for_update() does not have any effect.

        # we need to lock the selected records here, to make sure we do not
        # run this part in parallel (with the risk that when both transactions
        # are marking the run as finished, they still see each other as
        # unfinished since the transaction hasn't been committed yet).
        unfinished_siblings = len(instance.get_siblings().filter(
            return_dts__isnull=True).select_for_update())

        if instance.schedule_children and not unfinished_siblings:
            failed_siblings = instance.get_siblings().filter(
                return_success=False)

            if ((instance.return_success and not failed_siblings.count())
                    or job.schedule_children_on_error):
                for child in instance.job.children.all():
                    if child.enqueue_is_enabled:
                        child.schedule()



import json
import logging
import random
import time
from datetime import datetime, timedelta

import zmq
from django.conf import settings
from django.core.management.base import NoArgsCommand
from django.db import transaction

from job_runner.apps.job_runner.models import KillRequest, Run, Worker


logger = logging.getLogger(__name__)


class Command(NoArgsCommand):
    help = 'Broadcast runs and kill-requests to workers'

    publisher = None
    """
    Holds the ZMQ ``publisher`` instance, used to publish to the workers.
    """

    @transaction.commit_manually
    def handle_noargs(self, **options):
        logger.info('Starting queue broadcaster')
        context = zmq.Context(1)

        # setup the publisher to which the workers are subscribing
        self.publisher = context.socket(zmq.PUB)
        self.publisher.bind(
            'tcp://*:{0}'.format(settings.JOB_RUNNER_BROADCASTER_PORT))

        # give the subscribers some time to (re-)connect.
        time.sleep(2)

        ping_delta = timedelta(
            seconds=settings.JOB_RUNNER_WORKER_PING_INTERVAL)
        next_ping_request = datetime.utcnow()

        while True:
            if next_ping_request <= datetime.utcnow():
                self._broadcast_worker_ping()
                next_ping_request = datetime.utcnow() + ping_delta
            self._broadcast_runs()
            self._broadcast_kill_requests()
            transaction.commit()
            time.sleep(5)

        self.publisher.close()
        self.event_publisher.close()
        context.term()

    @transaction.commit_manually
    def _broadcast_runs(self):
        """
        Broadcast runs that are scheduled to run now.

        When the job has ``job__enqueue_is_enabled`` set to ``False``, its
        runs are not broadcasted, unless they are scheduled manually
        (``is_manual`` set to ``True``).

        """
        enqueueable_runs = Run.objects.enqueueable().select_related()

        # Use select_for_update so that the enqueueable runs will be locked.
        # This is to make sure that in case of multiple broadcasters we're not
        # creating any duplicate runs in case of run_on_all_workers=True.
        enqueueable_runs = enqueueable_runs.select_for_update()
        broadcasted = {}
        to_broadcast = []

        try:
            for run in enqueueable_runs:
                # schedule the run if we haven't already scheduled a run for
                # the same job, or when the run.schedule_id is equal to the
                # already scheduled run (which indicates that it needs to be
                # scheduled in parallel).
                if (run.job.pk not in broadcasted or
                        (run.job.pk in broadcasted and
                            broadcasted[run.job.pk] == run.get_schedule_id())):
                    worker = run.worker

                    # reschedule the job (the reschedule method will take care
                    # of checking if we should reschedule the job or not).
                    run.job.reschedule()

                    if not worker:
                        # if the job should run on all workers
                        if run.job.run_on_all_workers:
                            workers = run.job.worker_pool.workers.filter(
                                enqueue_is_enabled=True)
                            if workers.count():
                                broadcasted[run.job.pk] = run.get_schedule_id()

                                for w in workers:
                                    # assign a worker to each run
                                    assigned_run = Run.objects.create(
                                        job=run.job,
                                        schedule_id=run.get_schedule_id(),
                                        worker=w,
                                        schedule_dts=run.schedule_dts,
                                        is_manual=run.is_manual,
                                        schedule_children=run.schedule_children
                                    )
                                    to_broadcast.append((assigned_run, w))

                                # delete the "old" unassigned run
                                run.delete()

                        # select a random worker
                        else:
                            # TODO: take ping response into account?
                            workers = run.job.worker_pool.workers.filter(
                                enqueue_is_enabled=True)
                            if len(workers):
                                # pick a random active worker
                                worker = workers[
                                    random.randint(0, len(workers) - 1)]

                    # this is the case when a run has already a worker assigned
                    # to it, or when we selected a random worker.
                    if worker:
                        to_broadcast.append((run, worker))
                        broadcasted[run.job.pk] = run.get_schedule_id()

        except Exception:
            logger.exception('Something went wrong, rolling back transaction')
            transaction.rollback()
        else:
            transaction.commit()

            for brocast_args in to_broadcast:
                self._broadcast_run(*brocast_args)

    def _broadcast_run(self, run, worker):
        """
        Broadcast ``run`` to ``worker``.
        """
        message = [
            'master.broadcast.{0}'.format(worker.api_key),
            json.dumps({'run_id': run.id, 'action': 'enqueue'})
        ]
        logger.info('Sending: {0}'.format(message))
        self.publisher.send_multipart(message)

    def _broadcast_kill_requests(self):
        """
        Broadcast kill-requests.
        """
        kill_requests = KillRequest.objects.killable().select_related()

        for kill_request in kill_requests:
            run = kill_request.run
            worker = run.worker
            message = [
                'master.broadcast.{0}'.format(worker.api_key),
                json.dumps({
                    'kill_request_id': kill_request.id,
                    'action': 'kill',
                })
            ]
            logger.debug('Sending: {0}'.format(message))
            self.publisher.send_multipart(message)

    def _broadcast_worker_ping(self):
        """
        Broadcast ping-request to all the workers.
        """
        workers = Worker.objects.all()

        for worker in workers:
            message = [
                'master.broadcast.{0}'.format(worker.api_key),
                json.dumps({
                    'action': 'ping',
                })
            ]
            logger.debug('Sending: {0}'.format(message))
            self.publisher.send_multipart(message)



import os


PROJECT_ROOT = os.path.join(os.path.dirname(__file__), '..')

DEBUG = False
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@example.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.', # Add 'postgresql_psycopg2', 'mysql', 'sqlite3' or 'oracle'.
        'NAME': '',                      # Or path to database file if using sqlite3.
        'USER': '',                      # Not used with sqlite3.
        'PASSWORD': '',                  # Not used with sqlite3.
        'HOST': '',                      # Set to empty string for localhost. Not used with sqlite3.
        'PORT': '',                      # Set to empty string for default. Not used with sqlite3.
    }
}

# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# In a Windows environment this must be set to your system time zone.
TIME_ZONE = 'Europe/Amsterdam'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

# If you set this to False, Django will not format dates, numbers and
# calendars according to the current locale.
USE_L10N = True

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True

# Absolute filesystem path to the directory that will hold user-uploaded files.
# Example: "/home/media/media.lawrence.com/media/"
MEDIA_ROOT = os.path.join(PROJECT_ROOT, 'static', 'media')

# URL that handles the media served from MEDIA_ROOT. Make sure to use a
# trailing slash.
# Examples: "http://media.lawrence.com/media/", "http://example.com/media/"
MEDIA_URL = '/media/'

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
STATIC_ROOT = os.path.join(PROJECT_ROOT, 'static', 'static')

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
STATIC_URL = '/static/'

# Additional locations of static files
STATICFILES_DIRS = (
    os.path.join(PROJECT_ROOT, 'static', 'job_runner'),
)

# List of finder classes that know how to find static files in
# various locations.
STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
#    'django.contrib.staticfiles.finders.DefaultStorageFinder',
     'compressor.finders.CompressorFinder',
)

# Make this unique, and don't share it with anybody.
SECRET_KEY = '&amp;nghghm8)r1*_s$)p^+p==513yu9*aphzdt*qlc-jr(eyrd!u_'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
#     'django.template.loaders.eggs.Loader',
)

MIDDLEWARE_CLASSES = (
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.transaction.TransactionMiddleware',
    # Uncomment the next line for simple clickjacking protection:
    # 'django.middleware.clickjacking.XFrameOptionsMiddleware',
)

ROOT_URLCONF = 'job_runner.urls'

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = 'job_runner.wsgi.application'

TEMPLATE_DIRS = (
    os.path.join(PROJECT_ROOT, 'templates'),
)

INSTALLED_APPS = (
    # has to be before admin app
    'grappelli',

    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.staticfiles',

    'south',
    'tastypie',
    'compressor',
    'smart_selects',

    'job_runner.apps.job_runner',
)

# A sample logging configuration. The only tangible logging
# performed by this configuration is to send an email to
# the site admins on every HTTP 500 error when DEBUG=False.
# See http://docs.djangoproject.com/en/dev/topics/logging for
# more details on how to customize your logging configuration.
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            '()': 'django.utils.log.RequireDebugFalse'
        }
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ['require_debug_false'],
            'class': 'django.utils.log.AdminEmailHandler'
        }
    },
    'loggers': {
        'django.request': {
            'handlers': ['mail_admins'],
            'level': 'ERROR',
            'propagate': True,
        },
    }
}


LOGIN_REDIRECT_URL = '/'

GRAPPELLI_ADMIN_HEADLINE = 'Job-Runner'
GRAPPELLI_ADMIN_TITLE = 'Job-Runner'


API_LIMIT_PER_PAGE = 200


JOB_RUNNER_WS_SERVER = 'ws://localhost:5000/'
"""
The URL to the Job-Runner WebSocket server.

This should be in the following format::

    ws://hostname:port/

"""


JOB_RUNNER_ADMIN_EMAILS = []
"""
A list of e-mail addresses of the Job-Runner admin(s).

This list will currently be used when a job failed to reschedule.

"""


JOB_RUNNER_BROADCASTER_PORT = 5556
"""
The port to which the queue broadcaster is binding to.

Unless there is a specific need, you can keep the default.

"""


JOB_RUNNER_WS_SERVER_HOSTNAME = 'localhost'
"""
The hostname of the WebSocket Server.
"""


JOB_RUNNER_WS_SERVER_PORT = 5555
"""
The port of the WebSocket Server.
"""


JOB_RUNNER_WORKER_PING_INTERVAL = 60 * 5
"""
The interval in seconds for sending ping-requests to the workers.
"""


JOB_RUNNER_WORKER_HEALTH_CHECK_INTERVAL = 60 * 5
"""
The interval in seconds for running the health check.
"""


JOB_RUNNER_WORKER_PING_MARGIN = 15
"""
The time to add to the interval before considering a worker is not responding.

This is needed since the ping / pong are async (the PING is sent over ZMQ, the
pong is done by making a request to the REST API).

"""


JOB_RUNNER_WORKER_UNRESPONSIVE_AFTER_INTERVALS = 3
"""
The number of missed ping responses after which to declare a worker
unresponsive.
"""


HOSTNAME = ''
"""
The hostname of the server.

This value is used for generating URL's in the notification e-mails.

"""



import calendar
from datetime import timedelta

from django.conf import settings
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import signals
from django.template import Context, Template
from django.utils import timezone
from smart_selects.db_fields import ChainedForeignKey

from job_runner.apps.job_runner import notifications
from job_runner.apps.job_runner.managers import KillRequestManager, RunManager
from job_runner.apps.job_runner.signals import post_run_update, post_run_create
from job_runner.apps.job_runner.utils import correct_dst_difference


RESCHEDULE_INTERVAL_TYPE_CHOICES = (
    ('MINUTE', 'Every x minutes'),
    ('HOUR', 'Every x hours'),
    ('DAY', 'Every x days'),
    ('MONTH', 'Every x months'),
)


def validate_positive(value):
    """
    Validate that the value is positive (> 0).
    """
    if value <= 0:
        raise ValidationError(u'The input should be greater than 0.')


def validate_no_recursion(value, seen=[]):
    """
    Validate that there is no recursion in the job-chain.
    """
    # when the validate_no_recursion function is invoked, the ``value`` will
    # be the id of the object referencing to, so we have to lookup this object
    # first
    if isinstance(value, (int, long)):
        value = Job.objects.get(pk=value)

    if value in seen:
        raise ValidationError(u'Recursive loop of jobs detected.')
    seen.append(value)

    if value.parent:
        validate_no_recursion(value.parent, seen)

    return None


class RescheduleException(Exception):
    """
    Exception used when a rescheduling failed.
    """


class Project(models.Model):
    """
    Projects
    """
    title = models.CharField(max_length=255, db_index=True)
    description = models.TextField(blank=True)
    groups = models.ManyToManyField(
        Group,
        verbose_name='Viewers (groups)',
        help_text=(
            'These are the groups that can see the project in the dashboard. '
        )
    )
    auth_groups = models.ManyToManyField(
        Group,
        verbose_name='Project admins (groups)',
        blank=True,
        related_name='auth_groups_set',
        help_text=(
            'These are the groups that are authorized to see this project '
            'and its jobs in the admin and are able to re-schedule the jobs '
            'of this project in the dashboard.'
        )
    )
    worker_pools = models.ManyToManyField(
        'job_runner.WorkerPool',
        help_text=(
            'These are the worker-pools that will be available for this '
            'project.'
        )
    )
    notification_addresses = models.TextField(
        help_text='Separate e-mail addresses by a newline',
        blank=True,
    )
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, nothing for this project will be added to '
            'the worker queue. This will not affect already running jobs.'
        )
    )

    def __unicode__(self):
        return self.title

    def get_notification_addresses(self):
        """
        Return a ``list`` notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        return [x.strip() for x in addresses if x.strip() != '']

    class Meta:
        ordering = ('title', )


class Worker(models.Model):
    """
    Workers
    """
    title = models.CharField(max_length=255)
    description = models.TextField(blank=True)
    api_key = models.CharField(max_length=255, db_index=True, unique=True)
    secret = models.CharField(max_length=255, db_index=True)
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, nothing for this worker will be added to '
            'the worker queue. This will not affect already running jobs.'
        )
    )
    ping_response_dts = models.DateTimeField(
        blank=True, null=True, editable=False)

    # the worker will send back the worker version and concurrent jobs on
    # ping response
    worker_version = models.CharField(
        max_length=100, blank=True, null=True, editable=False)
    concurrent_jobs = models.PositiveIntegerField(
        blank=True, null=True, editable=False)

    def __unicode__(self):
        return self.title

    def is_responsive(self):
        """
        Return ``bool`` indicating if the worker is resposive.
        """
        unresponsive_intervals = \
            settings.JOB_RUNNER_WORKER_UNRESPONSIVE_AFTER_INTERVALS
        ping_interval = settings.JOB_RUNNER_WORKER_PING_INTERVAL
        ping_margin = settings.JOB_RUNNER_WORKER_PING_MARGIN

        acceptable_delta = timedelta(
            seconds=(unresponsive_intervals * ping_interval) + ping_margin)

        if self.ping_response_dts:
            if self.ping_response_dts + acceptable_delta >= timezone.now():
                return True

        return False

    class Meta:
        ordering = ('title', )


class WorkerPool(models.Model):
    """
    Worker-pool.
    """
    title = models.CharField(max_length=255, db_index=True)
    description = models.TextField(blank=True)
    notification_addresses = models.TextField(
        help_text='Separate e-mail addresses by a newline',
        blank=True,
    )
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, nothing for this worker-pool will be added to '
            'the worker queue. This will not affect already running jobs.'
        )
    )
    workers = models.ManyToManyField(Worker)

    def __unicode__(self):
        return self.title

    def get_notification_addresses(self):
        """
        Return a ``list`` notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        return [x.strip() for x in addresses if x.strip() != '']

    class Meta:
        ordering = ('title', )


class JobTemplate(models.Model):
    """
    Job templates
    """
    title = models.CharField(max_length=255, db_index=True)
    description = models.TextField(blank=True)
    body = models.TextField(help_text=(
        'Use {{ content|safe }} at the place where you want to render the '
        'script content of the job'
    ))
    project = models.ForeignKey(Project)
    notification_addresses = models.TextField(
        help_text='Separate e-mail addresses by a newline',
        blank=True,
    )
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, nothing for this template will be added to '
            'the worker queue. This will not affect already running jobs.'
        )
    )

    def __unicode__(self):
        return u'{0} > {1}'.format(self.project, self.title)

    def save(self, *args, **kwargs):
        """
        Override default save to re-save jobs.

        This will make sure that when the template changed, the cached version
        will be updated on the job level.

        """
        super(JobTemplate, self).save(*args, **kwargs)
        for job in self.job_set.all():
            job.save()

    def get_notification_addresses(self):
        """
        Return a ``list`` of notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        addresses = [x.strip() for x in addresses if x.strip() != '']
        addresses.extend(self.project.get_notification_addresses())
        return addresses

    class Meta:
        ordering = ('project__title', 'title', )


class Job(models.Model):
    """
    Contains the job data.
    """
    parent = models.ForeignKey(
        'self',
        blank=True,
        null=True,
        related_name='children',
        validators=[validate_no_recursion],
    )
    job_template = models.ForeignKey(JobTemplate)
    worker_pool = ChainedForeignKey(
        WorkerPool,
        chained_field='job_template',
        chained_model_field='project__jobtemplate',
        show_all=False,
        auto_choose=False,
        help_text='Select a job-template first to see the available pools.',
    )
    run_on_all_workers = models.BooleanField(
        default=False,
        help_text=(
            'Run this job on all workers within the selected pool. '
            'NOTE: be aware that the job will be duplicated to run in '
            'parallel on all workers of the pool. Leave this box unticked to '
            'have the job run once on only one of the workers.'
        )
    )
    schedule_children_on_error = models.BooleanField(
        default=False,
        help_text=(
            'Schedule children even when the job fails (or fails on one of '
            'the workers in case you ticked run on all workers). '
            'Normally a failed job means that it will stop the chain.'
        )
    )
    title = models.CharField(max_length=255, db_index=True)
    description = models.TextField(
        blank=True,
        help_text='You can use Markdown syntax to format this description.',
    )
    script_content_partial = models.TextField('script content')
    script_content = models.TextField(editable=False)
    enqueue_is_enabled = models.BooleanField(
        default=True,
        db_index=True,
        help_text=(
            'If unchecked, the job will not be added to the worker queue. '
            'This will not affect already running jobs.'
        )
    )
    reschedule_interval = models.PositiveIntegerField(
        null=True,
        blank=True,
        validators=[validate_positive],
        help_text=(
            'A positive number (greater than 0). '
            'Leave blank if you do not want to re-schedule this job.'
        )
    )
    reschedule_interval_type = models.CharField(
        max_length=6,
        blank=True,
        choices=RESCHEDULE_INTERVAL_TYPE_CHOICES,
        db_index=True,
        help_text=(
            'Note: with montly re-scheduling, the date will be incremented '
            'by the number of days in the month.'
        )
    )
    notification_addresses = models.TextField(
        help_text='Separate addresses by a newline',
        blank=True,
    )
    fail_times = models.PositiveIntegerField(
        editable=False,
        default=0,
        help_text='The number of times this job failed in a row.',
    )
    disable_enqueue_after_fails = models.PositiveIntegerField(
        blank=True,
        null=True,
        help_text=(
            'The number of times after which the enqueue-ing of a job should '
            'be disabled when it failed (blank = never disable enqeueue).'
        )
    )
    last_completed_schedule_id = models.PositiveIntegerField(
        blank=True,
        null=True,
        editable=False,
    )

    class Meta:
        ordering = (
            'job_template__project__title',
            'job_template__title',
            'title',
        )
        unique_together = (('title', 'job_template'),)

    def __unicode__(self):
        return u'{0} > {1}'.format(self.job_template, self.title)

    def save(self, *args, **kwargs):
        t = Template(self.job_template.body)
        c = Context({'content': self.script_content_partial})
        self.script_content = t.render(c)
        super(Job, self).save(*args, **kwargs)

    def get_notification_addresses(self):
        """
        Return a ``list`` of notification addresses.
        """
        addresses = self.notification_addresses.strip().split('\n')
        addresses = [x.strip() for x in addresses if x.strip() != '']
        addresses.extend(self.job_template.get_notification_addresses())
        addresses.extend(self.worker_pool.get_notification_addresses())
        return addresses

    def schedule(self, dts=None):
        """
        Schedule the job to run at the given ``dts``.

        :param dts:
            An instance of :class:`datetime.datetime` or ``None`` to schedule
            now.

        """
        if not dts:
            dts = timezone.now()

        Run.objects.create(
            job=self,
            schedule_dts=dts,
        )

    def reschedule(self):
        """
        Reschedule job.

        This will check if the job is setup for reschedule. Then it will
        try to get the next reschedule date. When getting this date fails
        (eg: when the rescheduling always falls within the reschedule exclude),
        it will send out an e-mail to the Job-Runner admins and the e-mail
        addresses that are setup for this job, script and server.

        """
        # there is already an other run which is not finished yet, do
        # not re-schedule, it will be rescheduled when the other job
        # finishes
        active_runs = self.run_set.filter(return_dts__isnull=True)
        active_schedule_ids = []

        # TODO: is there a way in the Django ORM to achieve the same?
        for active_run in active_runs:
            if not active_run.schedule_id in active_schedule_ids:
                active_schedule_ids.append(active_run.schedule_id)

        # since we are pre-scheduling (a new run is created, before the
        # scheduled run is sent to the worker), having one schedule id is valid
        if len(active_schedule_ids) > 1:
            return

        # check if job is setup for re-scheduling
        if self.reschedule_interval_type and self.reschedule_interval:
            try:
                # order by -pk to get the last non-manual scheduled run
                # from which we need to increment
                last_run = self.run_set.filter(
                    is_manual=False).order_by('-pk')[0]
            except IndexError:
                return

            try:
                reschedule_date = self._get_reschedule_date(
                    last_run.schedule_dts)

                # correct daylight saving-time changes to make sure we keep
                # re-scheduling at the same hour (in local time).
                reschedule_date = correct_dst_difference(
                    last_run.schedule_dts, reschedule_date)

                self.schedule(reschedule_date)
            except RescheduleException:
                notifications.reschedule_failed(self)

    def _get_reschedule_incremented_dts(self, increment_date):
        """
        Increment the given ``reference_date`` with the reschedule interval.

        :param increment_date:
            An instance of :class:`!datetime.datetime` to increment.

        :return:
            An instance of :class:`!datetime.datetime`.

        """
        if self.reschedule_interval_type == 'MINUTE':
            return increment_date + timedelta(minutes=self.reschedule_interval)
        elif self.reschedule_interval_type == 'HOUR':
            return increment_date + timedelta(hours=self.reschedule_interval)
        elif self.reschedule_interval_type == 'DAY':
            return increment_date + timedelta(days=self.reschedule_interval)
        elif self.reschedule_interval_type == 'MONTH':
            # increment the dts with the number of days that are in the
            # ``reschedule_date``.
            reschedule_date = increment_date
            for x in range(self.reschedule_interval):
                days_in_month = calendar.monthrange(
                    reschedule_date.year, reschedule_date.month)
                reschedule_date = reschedule_date + timedelta(
                    days=days_in_month[1])
            return reschedule_date

    def _get_reschedule_date(
            self,
            reference_date,
            increment_date=None):
        """
        Return a reschedule datetime.

        This will take the reschedule exclude times into account when
        generating a new date/time.

        :param reference_date:
            The reference :class:`datetime.datetime` to generate the
            reschedule date from.

        :param increment_date:
            The :class:`datetime.datetime` to increment for calculating the
            reschedule date. This is optional and when not set, the
            ``reference_date`` will be used.

        :raises:
            :exc:`.RescheduleException` when the reschedule date can not
            be calculated. This is for example the case when a reschedule
            exclude exists which covers the whole day.

        :return:
            An instance of :class:`datetime.datetime`.

        """
        # make sure the reference_date is in the current tz, because we are
        # going to compare hours / minutes!
        reference_date = reference_date.astimezone(
            timezone.get_default_timezone())

        if not increment_date:
            increment_date = reference_date

        reschedule_date = self._get_reschedule_incremented_dts(increment_date)

        while reschedule_date < timezone.now():
            reschedule_date = self._get_reschedule_incremented_dts(
                reschedule_date)

        if (increment_date - reference_date) > timedelta(days=1):
            raise RescheduleException(
                'Unable to reschedule due to reschedule excludes')

        for exclude in self.rescheduleexclude_set.all():
            if (reschedule_date.time() >= exclude.start_time and
                    reschedule_date.time() <= exclude.end_time):
                return self._get_reschedule_date(
                    reference_date, reschedule_date)

        return reschedule_date


class RescheduleExclude(models.Model):
    """
    Exclude rules for rescheduling.
    """
    job = models.ForeignKey(Job)
    start_time = models.TimeField()
    end_time = models.TimeField()
    note = models.CharField(max_length=255, blank=True)

    def __unicode__(self):
        return u'{0} - {1}'.format(self.start_time, self.end_time)


class Run(models.Model):
    """
    Contains the data related to a (scheduled) job run.
    """
    job = models.ForeignKey(Job)
    schedule_id = models.PositiveIntegerField(
        null=True, default=None, db_index=True)
    worker = models.ForeignKey(Worker, null=True, blank=True)
    schedule_dts = models.DateTimeField(db_index=True)
    enqueue_dts = models.DateTimeField(null=True, db_index=True)
    start_dts = models.DateTimeField(null=True, db_index=True)
    return_dts = models.DateTimeField(null=True, db_index=True)
    return_success = models.NullBooleanField(
        default=None, null=True, db_index=True)
    pid = models.PositiveIntegerField(null=True, default=None)
    is_manual = models.BooleanField(
        default=False, editable=False, db_index=True)
    schedule_children = models.BooleanField(
        default=True, editable=False, db_index=True)

    objects = RunManager()

    class Meta:
        ordering = (
            '-return_dts', '-start_dts', '-enqueue_dts', 'schedule_dts')

    @models.permalink
    def get_absolute_url(self):
        return ('job_runner:job_run', (), {
            'project_id': self.job.job_template.project.pk,
            'job_id': self.job.pk,
            'run_id': self.pk,
        })

    def get_siblings(self):
        """
        Return a ``QuerySet`` instance for the siblings of this run.
        """
        return self.__class__.objects.exclude(
            pk=self.pk,
            schedule_id__isnull=True,
        ).filter(
            schedule_id=self.schedule_id,
            job=self.job,
        )

    def get_schedule_id(self):
        if self.schedule_id:
            return self.schedule_id
        return self.pk

    def mark_failed(self, message):
        """
        Mark this run as failed (if ``return_dts`` is not set yet).

        :param message:
            A ``str`` explaining why this run was marked as failed.

        """
        if self.return_dts:
            return

        if not self.enqueue_dts:
            self.enqueue_dts = timezone.now()

        if not self.start_dts:
            self.start_dts = timezone.now()

        self.return_dts = timezone.now()
        self.return_success = False

        log_message = 'This run was marked as failed. Reason: {0}'.format(
            message)

        try:
            # In rare cases, it is possible that there is already a log for
            # the run.
            run_log = self.run_log
            run_log.content = log_message
            run_log.save()
        except RunLog.DoesNotExist:
            RunLog.objects.create(
                run=self,
                content=log_message
            )

        # saving the records triggers the sending of error e-mails so we need
        # to save the log before saving the run record.
        self.save()


class KillRequest(models.Model):
    """
    Contains requests to kill active runs.
    """
    run = models.ForeignKey(Run)
    schedule_dts = models.DateTimeField(auto_now_add=True, db_index=True)
    enqueue_dts = models.DateTimeField(null=True, db_index=True)
    execute_dts = models.DateTimeField(null=True, db_index=True)

    objects = KillRequestManager()


class RunLog(models.Model):
    """
    Contains log output for runs.
    """
    run = models.OneToOneField(Run, related_name='run_log')
    content = models.TextField(null=True, default=None)

    class Meta:
        ordering = ('-run',)


signals.post_save.connect(post_run_update, sender=Run)
signals.post_save.connect(post_run_create, sender=Run)



from setuptools import find_packages, setup

import job_runner


setup(
    name='job-runner',
    version=job_runner.__version__,
    url='http://github.com/spilgames/job-runner',
    license='BSD',
    author='Spil Games',
    author_email='datawarehouse@spilgames.com',
    description='Tool for scheduling jobs and realtime monitoring',
    long_description=open('README.rst').read(),
    packages=find_packages(),
    include_package_data=True,
    scripts=['manage.py'],
    install_requires=[
        'django==1.4.2',
        'django-grappelli',
        'django_compressor',
        'django-smart-selects',
        'django-tastypie==0.9.14',
        'south',
        'pytz',
        'pyzmq',
        'mimeparse==0.1.3',
        'python-dateutil==1.5',
    ],
)



from django.test import TestCase
from django.utils import timezone
from mock import Mock, call

from job_runner.apps.job_runner.management.commands.broadcast_queue import (
    Command)
from job_runner.apps.job_runner.models import (
    Job, JobTemplate, KillRequest, Project, Run, Worker, WorkerPool)


class CommandTestCase(TestCase):
    """
    Tests for :class:`.Command`.
    """
    fixtures = [
        'test_auth',
        'test_projects',
        'test_workers',
        'test_worker_pools',
        'test_job_templates',
        'test_jobs',
    ]

    def test__broadcast_kill_requests(self):
        """
        Test :meth:`.Command._broadcast_kill_requests`.
        """
        worker = Worker.objects.get(pk=1)
        run = Run.objects.get(pk=1)
        run.start_dts = timezone.now()
        run.pid = 1234
        run.worker = worker
        run.save()
        KillRequest.objects.create(
            run=run,
            schedule_dts=timezone.now()
        )

        command = Command()

        command.publisher = Mock()
        command._broadcast_kill_requests()

        self.assertEqual([
            call([
                'master.broadcast.worker1',
                '{"action": "kill", "kill_request_id": 1}'
            ])
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs(self):
        """
        Test :meth:`.Command._broadcast_runs`.
        """
        command = Command()

        command.publisher = Mock()
        command._broadcast_runs()

        self.assertEqual([
            call([
                'master.broadcast.worker1',
                '{"action": "enqueue", "run_id": 1}'
            ]),
            call([
                'master.broadcast.worker2',
                '{"action": "enqueue", "run_id": 2}'
            ]),
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs_run_on_all_workers(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``run_on_all_workers=True``.
        """
        job = Job.objects.get(pk=1)
        job.run_on_all_workers = True
        job.save()

        pool = WorkerPool.objects.get(pk=1)
        pool.workers.add(Worker.objects.get(pk=2))
        pool.save()

        Run.objects.get(pk=2).delete()

        command = Command()
        command.publisher = Mock()

        command._broadcast_runs()

        self.assertEqual([
            call([
                'master.broadcast.worker1',
                '{"action": "enqueue", "run_id": 3}'
            ]),
            call([
                'master.broadcast.worker2',
                '{"action": "enqueue", "run_id": 4}'
            ]),
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs_project_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.Project`.
        """
        Project.objects.update(enqueue_is_enabled=False)
        command = Command()

        command.publisher = Mock()
        command._broadcast_runs()

        self.assertEqual([
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs_worker_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.Worker`.
        """
        Worker.objects.update(enqueue_is_enabled=False)
        command = Command()

        command.publisher = Mock()
        command._broadcast_runs()

        self.assertEqual([
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs_job_template_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.JobTemplate`.
        """
        JobTemplate.objects.update(enqueue_is_enabled=False)
        command = Command()

        command.publisher = Mock()
        command._broadcast_runs()

        self.assertEqual([
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs_disabled_enqueue(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        for :class:`.Run`.
        """
        Job.objects.update(enqueue_is_enabled=False)
        command = Command()

        command.publisher = Mock()
        command._broadcast_runs()

        self.assertEqual([
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs_with_active_run(self):
        """
        Test :meth:`.Command._broadcast_runs` with still one active run.
        """
        Run.objects.create(
            job=Job.objects.get(pk=1),
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
            start_dts=timezone.now(),
        )
        Run.objects.create(
            job=Job.objects.get(pk=2),
            schedule_dts=timezone.now(),
            enqueue_dts=timezone.now(),
        )

        command = Command()
        command.publisher = Mock()
        command._broadcast_runs()

        self.assertEqual([], command.publisher.send_multipart.call_args_list)

    def test__broadcast_runs_disabled_enqueue_with_manual(self):
        """
        Test :meth:`.Command._broadcast_runs` with ``enqueue_is_enabled=False``
        but with manual one.
        """
        Job.objects.update(enqueue_is_enabled=False)
        Run.objects.filter(pk=2).update(is_manual=True)
        command = Command()

        command.publisher = Mock()
        command._broadcast_runs()

        self.assertEqual([
            call([
                'master.broadcast.worker2',
                '{"action": "enqueue", "run_id": 2}'
            ]),
        ], command.publisher.send_multipart.call_args_list)

    def test__broadcast_worker_ping(self):
        """
        Test :meth:`.Command._broadcast_worker_ping`.
        """
        command = Command()
        command.publisher = Mock()

        command._broadcast_worker_ping()

        self.assertEqual([
            call([
                'master.broadcast.worker1',
                '{"action": "ping"}'
            ]),
            call([
                'master.broadcast.worker2',
                '{"action": "ping"}'
            ]),
        ], command.publisher.send_multipart.call_args_list)
